import { get } from "svelte/store"
import { Component } from "./Component"
import { rowListUrl } from "../rowListScreen"
import { backendUiStore } from "builderStore"

export function makeLinkComponent(tableName) {
  return new Component("@budibase/standard-components/link")
    .normalStyle({
      color: "#757575",
      "text-transform": "capitalize",
    })
    .hoverStyle({
      color: "#4285f4",
    })
    .text(tableName)
    .customProps({
      url: `/${tableName.toLowerCase()}`,
      openInNewTab: false,
      color: "",
      hoverColor: "",
      underline: false,
      fontSize: "",
      fontFamily: "initial",
    })
}

export function makeMainForm() {
  return new Component("@budibase/standard-components/form")
    .type("div")
    .normalStyle({
      width: "700px",
      padding: "0px",
      "border-radius": "0.5rem",
      "box-shadow": "0 1px 2px 0 rgba(0, 0, 0, 0.05)",
      margin: "auto",
      "margin-top": "20px",
      "padding-top": "48px",
      "padding-bottom": "48px",
      "padding-right": "48px",
      "padding-left": "48px",
      "margin-bottom": "20px",
    })
    .instanceName("Form")
}

export function makeBreadcrumbContainer(tableName, text, capitalise = false) {
  const link = makeLinkComponent(tableName).instanceName("Back Link")

  const arrowText = new Component("@budibase/standard-components/text")
    .type("none")
    .normalStyle({
      "margin-right": "4px",
      "margin-left": "4px",
    })
    .text(">")
    .instanceName("Arrow")

  const textStyling = {
    color: "#000000",
  }
  if (capitalise) {
    textStyling["text-transform"] = "capitalize"
  }
  const identifierText = new Component("@budibase/standard-components/text")
    .type("none")
    .normalStyle(textStyling)
    .text(text)
    .instanceName("Identifier")

  return new Component("@budibase/standard-components/container")
    .type("div")
    .normalStyle({
      "font-size": "14px",
      color: "#757575",
    })
    .instanceName("Breadcrumbs")
    .addChild(link)
    .addChild(arrowText)
    .addChild(identifierText)
}

export function makeSaveButton(table, formId) {
  return new Component("@budibase/standard-components/button")
    .normalStyle({
      background: "#000000",
      "border-width": "0",
      "border-style": "None",
      color: "#fff",
      "font-family": "Inter",
      "font-weight": "500",
      "font-size": "14px",
      "margin-left": "16px",
    })
    .hoverStyle({
      background: "#4285f4",
    })
    .text("Save")
    .customProps({
      className: "",
      disabled: false,
      onClick: [
        {
          "##eventHandlerType": "Validate Form",
          parameters: {
            componentId: formId,
          },
        },
        {
          parameters: {
            providerId: formId,
          },
          "##eventHandlerType": "Save Row",
        },
        {
          parameters: {
            url: rowListUrl(table),
          },
          "##eventHandlerType": "Navigate To",
        },
      ],
    })
    .instanceName("Save Button")
}

export function makeTitleContainer(title) {
  const heading = new Component("@budibase/standard-components/heading")
    .normalStyle({
      margin: "0px",
      "margin-bottom": "0px",
      "margin-right": "0px",
      "margin-top": "0px",
      "margin-left": "0px",
      flex: "1 1 auto",
    })
    .type("h3")
    .instanceName("Title")
    .text(title)

  return new Component("@budibase/standard-components/container")
    .type("div")
    .normalStyle({
      display: "flex",
      "flex-direction": "row",
      "justify-content": "space-between",
      "align-items": "center",
      "margin-top": "32px",
      "margin-bottom": "32px",
    })
    .instanceName("Title Container")
    .addChild(heading)
}

const fieldTypeToComponentMap = {
  string: "stringfield",
  number: "numberfield",
  options: "optionsfield",
  boolean: "booleanfield",
  longform: "longformfield",
  datetime: "datetimefield",
  attachment: "attachmentfield",
  link: "relationshipfield",
}

export function makeTableFormComponents(tableId) {
  const tables = get(backendUiStore).tables
  const schema = tables.find(table => table._id === tableId)?.schema ?? {}
  return makeSchemaFormComponents(schema)
}

export function makeQueryFormComponents(queryId) {
  const queries = get(backendUiStore).queries
  const params = queries.find(query => query._id === queryId)?.parameters ?? []
  let schema = {}
  params.forEach(param => {
    schema[param.name] = { ...param, type: "string" }
  })
  return makeSchemaFormComponents(schema)
}

export function makeDatasourceFormComponents(datasource) {
  if (!datasource) {
    return []
  }
  return datasource.type === "table"
    ? makeTableFormComponents(datasource.tableId)
    : makeQueryFormComponents(datasource._id)
}

function makeSchemaFormComponents(schema) {
  let components = []
  let fields = Object.keys(schema || {})
  fields.forEach(field => {
    const fieldSchema = schema[field]
    const componentType = fieldTypeToComponentMap[fieldSchema.type]
    const fullComponentType = `@budibase/standard-components/${componentType}`
    if (componentType) {
      const component = new Component(fullComponentType)
        .instanceName(field)
        .customProps({
          field,
          label: field,
          placeholder: field,
        })
      if (fieldSchema.type === "options") {
        component.customProps({ placeholder: "Choose an option " })
      }
      if (fieldSchema.type === "boolean") {
        component.customProps({ text: field, label: "" })
      }
      components.push(component)
    }
  })
  return components
}
